﻿using System.Collections.Generic;
using System.Linq;
using StreamBase.SB;

namespace StreamBase.DataCache
{
    public static class DataCacheConstants
    {

        /* Stream names that are expected in a module */
        public static string RequestAllStateStream = "RequestAllState";  /* input is requestId */
        public static string StateTableStream = "StateTable";       /* output is key, priority, requestId, and data */
        public static string DeltaStream = "Delta";            /* output is key, isDel, priority, and data */

        /* Field names */
        public static string isDelField = "isDel";
        public static string keyField = "key";
        public static string priorityField = "priority";
        public static string requestIdField = "requestId";

        public static string dataField = "data";

        public static Schema requestSchema = new Schema(RequestAllStateStream, new Schema.Field[] {
            Schema.CreateField( CompleteDataType.ForInt(), requestIdField )
            });

        public static Schema deltaSchema = new Schema(DeltaStream, new Schema.Field[] {   // requires additional field dataField of type Tuple
			Schema.CreateField( CompleteDataType.ForBool(), isDelField ),
			Schema.CreateField( CompleteDataType.ForString(), keyField ),
			Schema.CreateField( CompleteDataType.ForLong(), priorityField )
            });

        public static Schema stateTableSchema = new Schema(StateTableStream, new Schema.Field[] { // requires additional dataField of type Tuple
			Schema.CreateField( CompleteDataType.ForString(), keyField ),
			Schema.CreateField( CompleteDataType.ForInt(), requestIdField ),
			Schema.CreateField( CompleteDataType.ForLong(), priorityField )
            });

        public static bool isSuperSet(Schema test, Schema requiredSet)
        {
            foreach (Schema.Field f in requiredSet.GetFields())
                if (!test.HasField(f.GetName()) || test.GetField(f.GetName()).GetDataType() != f.GetDataType())
                    return false;
            return true;
        }

        public static string getModulePath(string fullPath)
        {
            string[] parts = fullPath.Split('.');
            if (parts.Length > 1)
                return string.Join(".", parts, 0, parts.Length - 1);
            else
                return fullPath;
        }

        /* Fetch the set of modules that support this type of Data Cache! */
        public static string[] getDataCaches()
        {
            ISbClient sbm = SbClientManager.GetSbClientManager();

            HashSet<StreamProperties> iss = sbm.GetInputStreams();
            HashSet<StreamProperties> oss = sbm.GetOutputStreams();

            if (iss == null || oss == null)
                return null;

            // Start our basic set
            HashSet<string> icaches = new HashSet<string>();

            try
            {

                // Check for modules with request schemas
                foreach (StreamProperties s in iss)
                {
                    if (!s.GetName().Equals(RequestAllStateStream) || !isSuperSet(s.GetSchema(), requestSchema))
                        continue;

                    icaches.Add(getModulePath(s.GetQualifiedName()));
                }

                // Start our basic set
                HashSet<string> ocaches = new HashSet<string>();

                // Ensure that all of the modules have output streams
                foreach (StreamProperties s in oss)
                {

                    if (!s.GetName().Equals(DeltaStream) ||
                         !isSuperSet(s.GetSchema(), deltaSchema) ||
                         !s.GetSchema().HasField(dataField) ||
                         !(s.GetSchema().GetField(dataField).GetDataType() == DataType.TUPLE))
                        continue;

                    string module = getModulePath(s.GetQualifiedName());

                    if (!icaches.Contains(module))
                        continue;

                    ocaches.Add(module);
                }

                // Clear original cache
                icaches.Clear();

                // do again for the state table..
                foreach (StreamProperties s in oss)
                {
                    if (!s.GetName().Equals(StateTableStream) ||
                         !isSuperSet(s.GetSchema(), stateTableSchema) ||
                         !s.GetSchema().HasField(dataField) ||
                         !(s.GetSchema().GetField(dataField).GetDataType() == DataType.TUPLE))
                        continue;

                    string path = getModulePath(s.GetQualifiedName());

                    if (!ocaches.Contains(path))
                        continue;

                    icaches.Add(path);
                }
            }
            catch (TupleException)
            {
                return null;
            }

            return icaches.ToArray<string>();
        }
    }
}
