﻿using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Windows.Threading;
using StreamBase.SB;
using System;

namespace StreamBase.DataCache
{
    public delegate void newSchema(Schema schema);

    // Specialized version of an observable collection
    public class DataCacheModelCollection<T> : ObservableCollection<T>, IDataCacheConsumer
        where T : IDataCacheModel, new()
    {
        // Index for knowing where to put things
        private List<string> index;

        // Used for dispatching
        private pumpData updModel;
        private Dispatcher UIthread;

        // In case anyone else is monitoring the Schema
        public event newSchema SchemaHandler;
 
        // Initialize  (won't actually start until someone starts listening to data!)
        public DataCacheModelCollection(string cacheName)
            : this(cacheName, Dispatcher.CurrentDispatcher)
        {

        }

        public DataCacheModelCollection(string cacheName, Dispatcher uiThread)
        {
            this.UIthread = uiThread;
            updModel = new pumpData(updateModel);
            this.index = new List<string>();
            DataCacheService.StartDataCache(cacheName, this);
        }

        // Dispatch until later with the data -- this is critical!
        private delegate void pumpData(CacheUpdate[] upd, Schema schema, bool refresh);
        public void ConsumeDataCacheUpdate(string cacheName, CacheUpdate[] upd, bool refresh)
        {
            /*
            Trace.WriteLine("MODEL Dispatching CacheUpdate with elements " + upd.Length + " refresh " + refresh);
            UIthread.BeginInvoke(updModel, DispatcherPriority.ContextIdle, new object[] { upd, null, refresh });
             */
            updateModel(upd, null, refresh);
        }

        // Update (in UI thread) with the incoming data
        private void updateModel(CacheUpdate[] upd, Schema schema, bool refresh)
        {
            // Trace.WriteLine("MODEL Updating model in UI thread with " + ((upd!=null)?upd.Length.ToString():"0") + " updates, refresh " + refresh);

            if (schema != null)
            {
                //Trace.WriteLine("With schema: " + schema.ToString());
                if (SchemaHandler != null)
                    SchemaHandler(schema);
            }

            /*
            for (int i = 0; i < upd.Length; i++)
            {
                Trace.WriteLine("MODEL i " + i + ": " + upd[i].key + ", " + upd[i].isDel + ", " + upd[i].priority + ", " + upd[i].data.ToString());
            }
            */

            // If empty or data refresh, then clear everything out first
            if (Count == 0 || refresh)
            {
                //Trace.WriteLine("MODEL Initial refresh. Emptying out model.");
                Clear();
                index.Clear();
            }

            // Now load up all data!
            for (int i = 0; i < upd.Length; i++)
            {

                int idx = index.BinarySearch(upd[i].key);

                // If found -- then update or delete
                if (idx >= 0)
                {

                    // Removed row - delete row and indicate fire deletion!
                    if (upd[i].isDel)
                    {
                        index.RemoveAt(idx);
                        RemoveAt(idx);
                        //Trace.WriteLine("MODEL Removed row with key " + upd[i].key + " from idx " + idx);
                    }

                    // Otherwise this is an update
                    else
                    {
                        index[idx] = upd[i].key;
                        this[idx].UpdateFromTuple(upd[i].data);
                        //Trace.WriteLine("MODEL Updated row with key " + upd[i].key + " at idx " + idx);
                    }
                }

                // Doesn't exist in our cache..
                else
                {

                    // .. unless it is meant as a delete, in which case we have received a spurious delete.
                    if (upd[i].isDel)
                    {

                        // Spurious delete. Compression behind the scenes...
                        //Trace.WriteLine("MODEL Deleting of element that doesn't exist... ignoring.");
                        continue;
                    }

                    // Normal insert
                    else
                    {
                        int insertRow = ~idx;
                        //Trace.WriteLine("MODEL Inserting row with key " + upd[i].key + " at idx " + idx);
                        index.Insert(insertRow, upd[i].key);

                        // Insert new object into row!
                        T t = new T();
                        t.UpdateFromTuple(upd[i].data);
                        Insert(insertRow, t);
                    }
                }
            }
        }

        // Capture schema (for headers and column types)
        public void ConsumeDataCacheSchema(string cacheName, Schema schema)
        {
            // empty everything out!
            //Trace.WriteLine("MODEL Emptying things out... schema arrived. Should auto-generate code here.");
            try
            {
                /*
                UIthread.BeginInvoke(updModel, DispatcherPriority.ContextIdle, new object[] { new CacheUpdate[] { }, schema, true });
                 */
                updateModel( new CacheUpdate[] { }, schema, true);
                index.Clear();
            }
            catch (Exception e)
            {
                Trace.WriteLine("MODEL ERROR: " + e.ToString());
            }
            //Trace.WriteLine("MODEL Done emptying.");
        }

        public Dispatcher GetDispatcherThread()
        {
            return UIthread;
        }
    }
}
