﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using StreamBase.SB;
using System.Windows.Threading;
using System.Diagnostics;

namespace StreamBase.DataCache
{
    public class DataCacheService : IStreamConsumer, IDataCacheService
    {

        static private Random rand = new Random();

        // Public static methods - only public methods available!

        // Register as a listener for a particular cache
        static public IDataCacheService StartDataCache(string cacheName, IDataCacheConsumer consumer)
        {

            // Just start it up!
            return new DataCacheService(cacheName, consumer);
        }

        // State of initialization
        private enum State
        {
            STATE_INITIALIZING,
            STATE_LOADING,
            STATE_ACTIVE
        };

        private string path;
        private State cState;
        private StreamBase.SB.Tuple requestTuple;
        private Dictionary<string, CacheUpdate> data;
        private IDataCacheConsumer consumer;
        private Boolean foundStateStream;
        private Boolean foundDeltaStream;
        private Dispatcher dthread;
        private SendDataDelegate updModel;
        private bool pendingPurge;

        private DataCacheService(string path, IDataCacheConsumer consumer)
        {
            //Trace.WriteLine("DATACACHE Starting DataCache Service for " + path);

            this.path = path;
            this.data = new Dictionary<string, CacheUpdate>();
            this.consumer = consumer;
            this.requestTuple = DataCacheConstants.requestSchema.CreateTuple();
            this.req_requestIdIdx = DataCacheConstants.requestSchema.GetFieldIndex(DataCacheConstants.requestIdField);
            this.foundStateStream = false;
            this.foundDeltaStream = false;
            this.dthread = consumer.GetDispatcherThread();
            this.updModel = SendData;
            this.pendingPurge = false;

            // Start consuming the deltas
            SbClientManager.GetSbClientManager().SubscribeToStream(path + "." + DataCacheConstants.DeltaStream, this);

            // Start consuming snapshots
            SbClientManager.GetSbClientManager().SubscribeToStream(path + "." + DataCacheConstants.StateTableStream, this);

            SbClientManager.GetSbClientManager().Changed += HandleStatusChange;
        }

        private void HandleStatusChange(object sender, EventArgs e)
        {
            // something we changed.
            //Trace.WriteLine("DATACAHE STATUS Connected=" + ((SbClientManager)sender).GetConnected() + ", containers: " + ((SbClientManager)sender).GetContainers());
        }

        public void Shutdown()
        {
            // Unsubscribe to delta
            SbClientManager.GetSbClientManager().SubscribeToStream(path + "." + DataCacheConstants.DeltaStream, this);

            // Unsubscribe to state if not active yet
            if (cState != State.STATE_ACTIVE)
                SbClientManager.GetSbClientManager().UnsubscribeToStream(path + "." + DataCacheConstants.StateTableStream, this);
        }

        private int state_keyFieldIdx;
        private int state_requestIdIdx;
        private int state_priorityIdx;
        private int state_dataIdx;
        private int delta_keyFieldIdx;
        private int delta_priorityIdx;
        private int delta_isDelIdx;
        private int delta_dataIdx;
        private int req_requestIdIdx;
        private Schema dataSchema;

        public void StartNewSubscription(Schema schema, string StreamName)
        {
            string[] parts = StreamName.Split('.');
            string sName = parts[parts.Length - 1];

            //Trace.WriteLine("DATACACHE Starting new subscription for " + StreamName + " currently " + foundDeltaStream + " and " + foundStateStream);

            if (sName.Equals(DataCacheConstants.StateTableStream))
            {
                if (foundStateStream)
                {
                    foundDeltaStream = false;
                }
                else
                {
                    foundStateStream = true;
                }
                state_keyFieldIdx = schema.GetFieldIndex(DataCacheConstants.keyField);
                state_requestIdIdx = schema.GetFieldIndex(DataCacheConstants.requestIdField);
                state_priorityIdx = schema.GetFieldIndex(DataCacheConstants.priorityField);
                state_dataIdx = schema.GetFieldIndex(DataCacheConstants.dataField);
            }
            else if (sName.Equals(DataCacheConstants.DeltaStream))
            {
                if (foundDeltaStream)
                {
                    foundStateStream = false;
                }
                else
                {
                    foundDeltaStream = true;
                }
                delta_keyFieldIdx = schema.GetFieldIndex(DataCacheConstants.keyField);
                delta_isDelIdx = schema.GetFieldIndex(DataCacheConstants.isDelField);
                delta_priorityIdx = schema.GetFieldIndex(DataCacheConstants.priorityField);
                delta_dataIdx = schema.GetFieldIndex(DataCacheConstants.dataField);
                dataSchema = schema.GetField(delta_dataIdx).GetSchema();

                // Update schema for consumer
                if (dthread != null)
                {
                    dthread.BeginInvoke(updModel, DispatcherPriority.Normal, new object[] { false, this.dataSchema });
                }
                else
                {
                    updModel(false, this.dataSchema);
                }
            }

            // Initializing again!
            if (foundStateStream && foundDeltaStream)
            {

                // Start the data cache by subscribing to everything
                //Trace.WriteLine("DATACACHE Re-initializing cache state.");

                // Start again... request everything
                StartDataCache();
            }
        }

        // This may need some re-start capabilities around it -- start request again and again!
        private void StartDataCache()
        {

            // Set state to INITIALIZING
            this.cState = State.STATE_INITIALIZING;

            // Initalize request tuple
            try
            {
                this.requestTuple.SetInt(req_requestIdIdx, rand.Next());
            }
            catch (TupleException e)
            {
                Trace.WriteLine("DATACACHE  Fatal tuple exception, should never happen: " + e.ToString());
            }

            // Clear out cache
            lock (data)
            {
                data.Clear();
            }

            // Send off our request
            //Trace.WriteLine("DATACACHE Sending initial request token.");
            SbClientManager.GetSbClientManager().EnqueueTuple(path + "." + DataCacheConstants.RequestAllStateStream, requestTuple);
        }

        public void ConsumeTuple(StreamBase.SB.Tuple tuple, String streamName)
        {

            try
            {
                // Trace.WriteLine("DATACACHE Data cache received from stream " + streamName + " tuple " + tuple.ToString());

                // tuple is where the state management needs to go!
                if (streamName.Equals(DataCacheConstants.StateTableStream))
                {

                    // Get requestIds!
                    int requestId = tuple.GetInt(state_requestIdIdx);
                    int sentreqId = this.requestTuple.GetInt(req_requestIdIdx);

                    // Switch to loading if we were initializing and received token
                    if (cState == State.STATE_INITIALIZING && requestId == sentreqId)
                    {
                        cState = State.STATE_LOADING;
                        //Trace.WriteLine("DATACACHE Initialized. Now loading.");
                        return;
                    }

                    // If loading and received a state update, then apply to model
                    if (cState == State.STATE_LOADING && requestId == sentreqId)
                    {
                        ApplyTupleToModel(tuple.GetString(state_keyFieldIdx),
                                tuple.GetLong(state_priorityIdx), false,
                                tuple.GetTuple(state_dataIdx));
                        return;
                    }

                    // If loading and received end token, then switch to active and purge
                    if (cState == State.STATE_LOADING && requestId == (-1 * sentreqId))
                    {
                        //Trace.WriteLine("DATACACHE Loaded. Now active.");
                        cState = State.STATE_ACTIVE;

                        // Unsubscribe to state stream!
                        SbClientManager.GetSbClientManager().UnsubscribeToStream(path + "." + DataCacheConstants.StateTableStream, this);

                        /* Before purging, let's delete elements that are marked for deletion - this is for convenience only */
                        lock (data)
                        {
                            List<string> removeKeys = new List<string>();
                            foreach (string key in data.Keys)
                                if (data[key].isDel)
                                    removeKeys.Add(key);

                            foreach (string key in removeKeys)
                                data.Remove(key);
                        }

                        /* Purge model to the downstream system */
                        PurgeModel(true);
                        return;
                    }
                }

                else if (streamName.Equals(DataCacheConstants.DeltaStream))
                {

                    // Simply update the model!
                    ApplyTupleToModel(tuple.GetString(delta_keyFieldIdx),
                            tuple.GetLong(delta_priorityIdx), tuple.GetBoolean(delta_isDelIdx),
                            tuple.GetTuple(delta_dataIdx));
                }
            }
            catch (TupleException e)
            {
                Trace.WriteLine("DATACACHE Fatal error parsing tuples, should never happen: " + e.ToString());
            }
        }

        // Apply a delta (whether original state or subsequent delta) to our internal data cache
        private void ApplyTupleToModel(string key, long priority, Boolean isDel, StreamBase.SB.Tuple tuple)
        {
            lock (data)
            {
                // If the data is already in our cached data model and more up to date, then skip it
                if (data.ContainsKey(key) && data[key].priority > priority)
                    return;

                // As this is not in our model or more up to do, then just set it
                CacheUpdate dd = new CacheUpdate();
                dd.key = key;
                dd.priority = priority;
                dd.data = tuple;
                dd.isDel = isDel;
                if (data.ContainsKey(key))
                    data[key] = dd;
                else
                    data.Add(key, dd);
                return;
            }
        }

        // Purge our data model if it isn't empty, i.e., there are things to push downstream
        // AND we are active, then purge the model
        public void NotifyComplete()
        {
            bool isData = false;
            lock (data)
            {
                isData = (data.Count > 0 && cState == State.STATE_ACTIVE);
            }
            
            if (isData) {
                PurgeModel(false);
            }
        }

        private void PurgeModel(bool refresh)
        {
            // Dispatch it if requested
            if (dthread != null)
            {
                // Use the same data lock..
                lock (data)
                {
                    // skip it if a purge is pending
                    if (pendingPurge)
                        return;

                    // Set it to true
                    pendingPurge = true;
                }

                //Trace.WriteLine("BEGIN INVOKE BEGIN INVOKE");
                dthread.BeginInvoke(updModel, DispatcherPriority.ApplicationIdle, new object[] { refresh, null });
                return;
            }

            // Otherwise call it directly
            SendData(refresh,null);
        }

        private delegate void SendDataDelegate(bool refresh, Schema scm);
        private void SendData(bool refresh, Schema scm)
        {
            if (scm == null)
            {
                CacheUpdate[] outarr;
                lock (data)
                {
                    // Skip if no data to send and refresh is false!
                    if (data.Count == 0 && !refresh)
                        return;

                    outarr = new CacheUpdate[data.Values.Count];
                    data.Values.CopyTo(outarr, 0);
                    data.Clear();

                    // No more pending purge
                    pendingPurge = false;
                }

                //Trace.WriteLine("DATACACHE Purging values with refresh " + refresh + " values " + outarr.Length);

                // Send off cache to downstream consumers
                consumer.ConsumeDataCacheUpdate(this.path, outarr, refresh);
            }
            else
            {
                consumer.ConsumeDataCacheSchema(this.path, scm);
            }
        }
    }
}
