﻿using System;
using System.Diagnostics;
using System.Threading;
using System.Collections.Generic;
using System.Runtime.CompilerServices;
using System.Linq;
using System.Text;
using StreamBase.SB;
using StreamBase.SB.Client;

namespace StreamBase.DataCache
{
    // Class for the singleton
    public class SbClientManager : ISbClient
    {

        // Singeton
        private static SbClientManager _singleton = null;

        // Get the singleton
        public static ISbClient GetSbClientManager()
        {
            if (_singleton == null)
                _singleton = new SbClientManager();
            return _singleton;
        }

        // Public event for finding out change in status
        public event ChangedStatusHandler Changed;

        //
        // Private state!
        //
        private static Boolean _running = true;                                     // Client is running or not
        private static TimeSpan _updateDelay;                                       // 50ms time loop
        private static Int32 _cyclesPerManagementOperation = 4;                     // Cycles per major operation (connect attempt, status check)
        private StreamBaseClient _sbc;                                              // Client
        private StreamBaseURI _uri;                                                 // Actual URI
        private Boolean _needsConnecting;                                           // Whether a connection is needed or not
        private String[] _containers;                                               // Current set of containers
        private Dictionary<String, HashSet<IStreamConsumer>> _consumers;            // Current set of consumers
        private Dictionary<String, HashSet<IStreamConsumer>> _subscribingConsumers; // Hopeful set of consumers

        // Private constructers
        private SbClientManager()
        {
            //Trace.WriteLine("Constructing SbClientManager.");

            // Load default settings from config
            _updateDelay = new TimeSpan(0, 0, 0, 0, 50);
            _cyclesPerManagementOperation = 20;

            // Setup consumers
            _consumers = new Dictionary<string, HashSet<IStreamConsumer>>();
            _subscribingConsumers = new Dictionary<string, HashSet<IStreamConsumer>>();

            // We need connecting..
            _needsConnecting = true;

            // Empty container list
            _containers = new string[0];

            // Start thread! This will automatically start connecting.. same logic for re-connections as initial!
            // (initial connection just has an empty list of subscribers.. maybe)
            new Thread(new ThreadStart(BackgroundThread)).Start();
        }

        private void BackgroundThread()
        {
            long cycles = 0;

            //Trace.WriteLine("background thread starting.");

            try
            {
                // For as long as we aren't shut down...
                while (_running)
                {
                    // Reset on cycles per management operation
                    if (cycles == _cyclesPerManagementOperation)
                        cycles = 0;

                    // Connect if needed
                    if (_needsConnecting)
                    {
                        // If connecting, then repeat! But only on a proper cycle.
                        if (cycles == 0 && TryToConnect())
                        {
                            UpdateStatus();
                            _needsConnecting = false;
                        }
                        continue;
                    }

                    // Otherwise we're connected..
                    else
                    {

                        // If closed, then go back to needsConnecting state
                        if (_sbc.IsClosed())
                        {
                            _needsConnecting = true;
                            UpdateStatus();
                            continue;
                        }

                        // Try patient subscriptions on a slower cycle
                        if (cycles == 0)
                        {
                            TryPatientSubscriptions();  // Hopeful subscribers (only kind)
                            UpdateContainers();         // Update list of containers
                        }

                        // If there are any active subscriptions, then try to dequeue
                        if (_consumers.Count > 0)
                            TryDequeue();
                    }

                    // Another cycle
                    cycles++;

                    // Sleep for a brief delay
                    try
                    {
                        Thread.Sleep(_updateDelay);
                    }
                    catch (ThreadInterruptedException)
                    {
                        /* */
                    }
                }
            }
            catch (Exception e)
            {
                Trace.WriteLine("FATAL: ERROR IN SBCLIENT: " + e.ToString());
            }

            //Trace.WriteLine("Background thread exiting.");
        }

        public void SetSBURI(string u)
        {
            try
            {
                // Get the URI
                _uri = new StreamBaseURI(u);
            }
            catch (URIException e)
            {
                Trace.WriteLine("Failed parsing URI: " + e.ToString());
            }
        }

        private void UpdateContainers()
        {
            if (!GetConnected())
            {
                _containers = new string[0];
                return;
            }

            string[] ncontainers = null;
            try
            {
                lock (_sbc)
                {
                    ncontainers = _sbc.ListEntities(EntityType.CONTAINER);
                }
            }
            catch (Exception e)
            {
                Trace.WriteLine("Failed getting containers: " + e.ToString());
                _sbc.Close();
            }

            if (ncontainers == null)
                ncontainers = new string[0];

            // Assume not updated
            bool upd = false;

            // Different size set means updated!
            if (_containers == null || _containers.Length != ncontainers.Length)
            {
                upd = true;
            }

            // Otherwise check all elements - if any different, then updated
            else
            {
                for (int i = 0; i < _containers.Length; i++)
                {
                    if (!_containers[i].Equals(ncontainers[i]))
                    {
                        upd = true;
                        break;
                    }
                }
            }

            // Not updated- return
            if (!upd)
                return;

            // Update it and notify
            _containers = ncontainers;
            UpdateStatus();
        }

        private void UpdateStatus()
        {
            //Trace.WriteLine("Updating status = " + getConnected() + " containers: " + getContainers());
            if (Changed != null)
                Changed(this, null);
        }

        public string[] GetContainers()
        {
            return _containers;
        }

        public bool GetConnected()
        {
            if (_sbc == null)
                return false;

            lock (_sbc)
            {
                if (!_sbc.IsClosed())
                    return true;
                else
                    return false;
            }
        }

        public void Shutdown()
        {
            _running = false;
        }

        public void Enqueue(string streamName, StreamBase.SB.Tuple tuple)
        {
            // @TODO This code should be defensive...

            lock (_sbc)
            {
                _sbc.Enqueue(streamName, tuple);
            }
        }

        [MethodImpl(MethodImplOptions.Synchronized)]
        public HashSet<StreamProperties> GetInputStreams()
        {
            if (!GetConnected())
                return null;

            try
            {
                lock (_sbc)
                {
                    HashSet<StreamProperties> props = new HashSet<StreamProperties>();
                    foreach (string c in _sbc.ListEntities(EntityType.CONTAINER))
                    {
                        foreach (string s in _sbc.ListEntities(c + "." + EntityType.INPUT_STREAMS, StreamBaseClient.ListEntitiesFlags.FULLY_QUALIFIED_NAMES))
                        {
                            props.Add(_sbc.GetStreamProperties(s));
                        }
                    }
                    /*
                    string[] inputStreams = sbc.ListEntities(EntityType.INPUT_STREAMS);
                    for (int i = 0; i < inputStreams.Length; i++)
                        props.Add(sbc.GetStreamProperties(inputStreams[i]));
                     */
                    return props;
                }
            }
            catch (StreamBaseException e)
            {
                Trace.WriteLine("Exception getting input streams: " + e.ToString());
                return null;
            }
        }

        public HashSet<StreamProperties> GetOutputStreams()
        {
            if (!GetConnected())
                return null;

		    try {
                lock (_sbc)
                {
                    HashSet<StreamProperties> props = new HashSet<StreamProperties>();
                    foreach (string c in _sbc.ListEntities(EntityType.CONTAINER))
                    {
                        foreach (string s in _sbc.ListEntities(c + "." + EntityType.OUTPUT_STREAMS, StreamBaseClient.ListEntitiesFlags.FULLY_QUALIFIED_NAMES))
                        {
                            props.Add(_sbc.GetStreamProperties(s));
                        }
                    }
                    /*
                    string[] outputStreams = sbc.ListEntities(EntityType.OUTPUT_STREAMS);
                    HashSet<StreamProperties> props = new HashSet<StreamProperties>();
                    for (int i = 0; i < outputStreams.Length; i++)
                        props.Add(sbc.GetStreamProperties(outputStreams[i]));
                     */
                    return props;
                }
            }
            catch (StreamBaseException e)
            {
                Trace.WriteLine("Exception getting output streams: " + e.ToString());
                return null;
            }
        }

        public bool EnqueueTuple(string stream, StreamBase.SB.Tuple data)
        {
            try
            {
                lock (_sbc)
                {
                    _sbc.Enqueue(stream, data);
                    return true;
                }
            }
            catch (StreamBaseException e)
            {
                Trace.WriteLine("Error sending: " + e.ToString());
                return false;
            }
        }

        public void UnsubscribeToStream(string streamName, IStreamConsumer cons)
        {
            Trace.WriteLine("Unsubscribing from stream " + streamName + " for a consumer");

            /* doesn't quite work yet -- conflict -- ignoring unsubscriptions for now */
            /* Problem seems to be that while a tuple is being consumed someone may try to */
            /* unsubscribe.. namely after the end of the StateTable, when finished, they may */
            /* unsubscribe immediately! */
            /*
            lock (subscribingConsumers)
            {
                if (subscribingConsumers.ContainsKey(streamName))
                    subscribingConsumers[streamName].Remove(cons);
            }

            lock (sbc) {
                if (consumers.ContainsKey(streamName))
                {
                    consumers[streamName].Remove(cons);
                    if (consumers[streamName].Count == 0)
                    {
                        try
                        {
                            sbc.Unsubscribe(streamName);
                        }
                        catch (Exception e)
                        {
                            Trace.WriteLine("Failed unsubscribing to " + streamName + ": " + e.ToString());
                        }
                    }
                }
            }
            */
        }

        // Subscribe in a patient way (keep trying until it's done!)
        public void SubscribeToStream(string streamName, IStreamConsumer cons)
        {
            lock (_subscribingConsumers)
            {
                if (!_subscribingConsumers.ContainsKey(streamName))
                    _subscribingConsumers.Add(streamName, new HashSet<IStreamConsumer>());

                _subscribingConsumers[streamName].Add(cons);
            }
        }

        private bool TryToConnect()
        {
            if (_uri == null)
                return false;

            if (GetConnected())
                Trace.WriteLine("FATAL ERROR: Connecting when already connected.");

            try
            {
                Trace.WriteLine("Connecting to StreamBase at " + _uri.ToString());
                _sbc = new StreamBaseClient(_uri);
            }
            catch (StreamBaseException)
            {
                Trace.WriteLine("Connection failed.");
                return false;
            }

            // Take a copy of the consumers
            if (_consumers.Count > 0)
            {
                //Trace.WriteLine("Resubscribing to " + _consumers.Count + " streams.");

                // Subscribe to the streams all a-fresh!
                foreach (KeyValuePair<String, HashSet<IStreamConsumer>> kvp in _consumers)
                {
                    foreach (IStreamConsumer isc in kvp.Value)
                    {
                        SubscribeToStream(kvp.Key, isc);
                    }
                }

                // Clear out current subscriptions
                _consumers.Clear();
            }

            //Trace.WriteLine("Done Connecting");

            // Succeeded in connecting
            return true;
        }


        private void TryPatientSubscriptions()
        {
            lock (_subscribingConsumers)
            {
                List<string> empty = new List<string>();
                foreach (string streamName in _subscribingConsumers.Keys)
                {
                    List<IStreamConsumer> subscribed = new List<IStreamConsumer>();
                    foreach (IStreamConsumer isc in _subscribingConsumers[streamName])
                    {
                        //Trace.WriteLine("Trying to subscribe to " + streamName);
                        if (RealSubscribeToStream(streamName, isc))
                        {
                            //Trace.WriteLine("Subscribed to " + streamName);
                            subscribed.Add(isc);
                        }
                        else
                        {
                            //Trace.WriteLine("Failed subscribing to " + streamName);
                        }
                    }
                    foreach (IStreamConsumer isc in subscribed)
                    {
                        _subscribingConsumers[streamName].Remove(isc);
                    }
                    if (_subscribingConsumers.Count == 0)
                        empty.Add(streamName);
                }

                foreach (string s in empty)
                {
                    _subscribingConsumers.Remove(s);
                }
            }
        }

        private bool RealSubscribeToStream(string streamName, IStreamConsumer iStreamConsumer)
        {
            // Send them the schema that they've just subscribed to
            lock (_sbc)
            {
                Schema streamSchema;
                try
                {
                    streamSchema = _sbc.GetSchemaForStream(streamName);
                }
                catch (StreamBaseException e)
                {
                    Trace.WriteLine("Failed subscribing to " + streamName + " : " + e.ToString());
                    return false;
                }

                // If not subscribed to this stream yet, then do so!
                if (!_consumers.ContainsKey(streamName))
                {
                    try
                    {
                        _sbc.Subscribe(streamName);
                    }
                    catch (StreamBaseException e)
                    {
                        Trace.WriteLine("StreamBaseException: " + e.ToString());
                        Trace.WriteLine("Stack: " + e.StackTrace);
                        return false;
                    }
                    _consumers.Add(streamName, new HashSet<IStreamConsumer>());
                }

                // Add in this particular consumer
                _consumers[streamName].Add(iStreamConsumer);

                // Start a new subscription
                iStreamConsumer.StartNewSubscription(streamSchema, streamName);

                return true;
            }
        }

        private void TryDequeue()
        {
            try
            {
                // If we can dequeue, then do it!
                while (true)
                {
                    DequeueResult dr;
                    lock (_sbc)
                    {
                        dr = _sbc.Dequeue(2);
                    }

                    if (dr.GetTupleCount() == 0)
                        break;

                    DispatchResult(dr);
                }
            }
            catch (StreamBaseException e)
            {
                Trace.WriteLine("StreamBaseException during dequeue: "
                    + e.ToString());
            }
        }

        private void DispatchResult(DequeueResult dr)
        {
            HashSet<IStreamConsumer> toNotify = new HashSet<IStreamConsumer>();

            foreach (StreamBase.SB.Tuple t in dr)
            {

                //Trace.WriteLine("Got tuple from stream " + dr.GetStreamProperties().GetQualifiedName() + " datA: " + t.ToString());
                HashSet<IStreamConsumer> hs;

                if (_consumers.TryGetValue(dr.GetStreamProperties().GetQualifiedName(), out hs))
                {
                    foreach (IStreamConsumer isc in hs)
                    {
                        //Trace.WriteLine("SEnding tuple " + t.ToString());
                        SendTuple(t, dr.GetStreamProperties().GetQualifiedName(), isc);
                        isc.ConsumeTuple(t, dr.GetStreamName());
                        toNotify.Add(isc);
                    }
                }

                if (_consumers.TryGetValue(dr.GetStreamName(), out hs))
                {
                    foreach (IStreamConsumer isc in hs)
                    {
                        //Trace.WriteLine("SEnding tuple " + t.ToString());
                        SendTuple(t, dr.GetStreamName(), isc);
                        toNotify.Add(isc);
                    }
                }
            }

            // this needs to be a bit more precise! only for those who wanted this stream..
            foreach (IStreamConsumer isc in toNotify)
            {
                //Trace.WriteLine("notifying complete.");
                isc.NotifyComplete();
            }
        }

        private void SendTuple(StreamBase.SB.Tuple t, string streamName, IStreamConsumer isc)
        {
            try
            {
                isc.ConsumeTuple(t, streamName);
            }
            catch (Exception e)
            {
                Trace.WriteLine("Error consuming tuple: " + e.ToString());
            }
        }
    }
}
