package com.streambase.contrib.datacache;

import java.util.Set;

import com.streambase.sb.DataType;
import com.streambase.sb.Schema;
import com.streambase.sb.StreamProperties;
import com.streambase.sb.TupleException;

/**
 * Internal constants used by the framework
 */
/* package */ class DataCacheConstants {
	
	/* Stream names that are expected in a module */
	static String RequestAllStateStream = new String( "RequestAllState" );  /* input is requestId */
	static String StateTableStream      = new String( "StateTable" );       /* output is key, priority, requestId, and data */
	static String DeltaStream           = new String( "Delta" );            /* output is key, isDel, priority, and data */

	/* Field names */
	static String isDelField            = new String( "isDel" );
	static String keyField              = new String( "key" );
	static String priorityField         = new String( "priority" );
	static String requestIdField        = new String( "requestId" );
	static String dataField             = new String( "data" );
	
	static Schema requestSchema = new Schema( RequestAllStateStream,
			Schema.createField( DataType.INT, requestIdField ) );

	static Schema deltaSchema = new Schema( DeltaStream,    // requires additional field dataField of type Tuple
			Schema.createField( DataType.BOOL, isDelField ),
			Schema.createField( DataType.STRING, keyField ),
			Schema.createField( DataType.LONG, priorityField ) );

	static Schema stateTableSchema = new Schema( StateTableStream, // requires additional dataField of type Tuple
			Schema.createField( DataType.STRING, keyField ),
			Schema.createField( DataType.INT, requestIdField ),
			Schema.createField( DataType.LONG, priorityField ) );
	
	/* Fetch the set of modules that support this type of Data Cache! */
	public static Set< String > getDataCaches() {
		
		DataCacheStreamBaseClient sbm = DataCacheStreamBaseClient.getClient();

		Set< StreamProperties > is = sbm.getInputStreams();
		Set< StreamProperties > os = sbm.getOutputStreams();
		
		if ( is == null || os == null )
			return null;

		// Start our basic set
		Set< String > icaches = new java.util.HashSet<String>();

		try {
	
			// Check for modules with request schemas
			for ( StreamProperties s : is ) {
				if ( s.getName().equals( RequestAllStateStream ) &&
						s.getSchema().isSupersetOf( requestSchema ) ) {
					icaches.add( s.getPath().getAllButLastComponent().toString() );
				}
			}
	
			// Start our basic set
			Set< String > ocaches = new java.util.HashSet<String>();
	
			// Ensure that all of the modules have output streams
			for ( StreamProperties s : os ) {
				if ( s.getName().equals( DeltaStream ) &&
					 s.getSchema().isSupersetOf( deltaSchema ) &&
					 s.getSchema().hasField( dataField ) &&
					 s.getSchema().getField( dataField ).getDataType() == DataType.TUPLE &&
					 icaches.contains( s.getPath().getAllButLastComponent().toString() )) {
					ocaches.add( s.getPath().getAllButLastComponent().toString() );
				}
			}
			
			// Clear original cache
			icaches.clear();
			
			for ( StreamProperties s : os ) {
				if ( s.getName().equals( StateTableStream ) &&
					 s.getSchema().isSupersetOf( stateTableSchema ) &&
					 s.getSchema().hasField( dataField ) &&
					 s.getSchema().getField( dataField ).getDataType() == DataType.TUPLE &&
					 ocaches.contains( s.getPath().getAllButLastComponent().toString() )) {
					icaches.add( s.getPath().getAllButLastComponent().toString() );
				}
			}
		}
		catch ( TupleException e ) {
			return null;
		}
		
		return icaches;
	}
}
