package com.streambase.contrib.datacache;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import com.streambase.contrib.datacache.DataCacheStreamBaseClient.IStreamConsumer;
import com.streambase.sb.Schema;
import com.streambase.sb.Tuple;
import com.streambase.sb.TupleException;

/**
 * Primary entry point to the data cache framework.
 * Users should retrieve the set of available data caches by invoking {@link #getDataCaches()},
 * and may then call {@link #startDataCache(String, IDataCacheConsumer)} to connect to a cache.
 * <p/>
 * This DataCacheService uses {@link DataCacheStreamBaseClient} for connectivity to a streambase
 * application, and users of a data cache should ensure they have connected to a streambase
 * application by calling {@link DataCacheStreamBaseClient#initClient(String[])}
 * 
 * @author mark.scannell@streambase.com
 */
public class DataCacheService implements IStreamConsumer {
	
	// Private static management members 
	static private Set< String >               cacheNames = DataCacheConstants.getDataCaches();
	
	// Public static methods - only public methods available!
	
	/**
	 * Connects an {@link IDataCacheConsumer} to a data cache.
	 * @param cacheName Data cache name. Use {@link #getDataCaches()} to know what names are valid. 
	 * @param consumer a listener that will be notified to maintain a view over the target data cache query table.
	 */
	static public void startDataCache( String cacheName, IDataCacheConsumer consumer ) {
	
		// Just start it up!
		new DataCacheService( cacheName, consumer );
	}


	/**
	 * Retrieves the set of available Data Cache names. This set is initialized
	 * upon first use of the data cache framework, and does not change.
	 * @return The names of data caches found when this service was loaded.
	 * Use these names when invoking {@link #startDataCache(String, IDataCacheConsumer)}.
	 */
	static public Set< String > getDataCaches() {
		return cacheNames;
	}
	
	// Random number generating requestIds
	static private Random rand = new Random();
	
	// State of initialization
	private enum State {
		STATE_INITIALIZING,
		STATE_LOADING,
		STATE_ACTIVE
	};
	
	private String path;
	private State  cState;
	private Tuple  requestTuple;
	private Map< String, IDataCacheConsumer.DataObj > data;
	private IDataCacheConsumer consumer;

	private DataCacheService( String path, IDataCacheConsumer consumer ) {
		this.path = path;
		this.data = new HashMap<String, IDataCacheConsumer.DataObj >();
		this.consumer = consumer;
		this.requestTuple = DataCacheConstants.requestSchema.createTuple();
		this.req_requestIdIdx = DataCacheConstants.requestSchema.getFieldIndex( DataCacheConstants.requestIdField );

		// Start the data cache by subscribing to everything
		startDataCache();
	}

	private int state_keyFieldIdx;
	private int state_requestIdIdx;
	private int state_priorityIdx;
	private int state_dataIdx;
	private int delta_keyFieldIdx;
	private int delta_priorityIdx;
	private int delta_isDelIdx;
	private int delta_dataIdx;
	private int req_requestIdIdx;
	private Schema dataSchema;

	@Override
	public void consumeSchema(Schema schema, String StreamName) {
		String parts[] = StreamName.split( "\\." );
		String sName = parts[ parts.length - 1 ];
		
		if ( sName.equals( DataCacheConstants.StateTableStream ) ) {
			state_keyFieldIdx  = schema.getFieldIndex( DataCacheConstants.keyField );
			state_requestIdIdx = schema.getFieldIndex( DataCacheConstants.requestIdField );
			state_priorityIdx  = schema.getFieldIndex( DataCacheConstants.priorityField );
			state_dataIdx      = schema.getFieldIndex( DataCacheConstants.dataField );
		}
		else if ( sName.equals( DataCacheConstants.DeltaStream ) ) {
			delta_keyFieldIdx  = schema.getFieldIndex( DataCacheConstants.keyField );
			delta_isDelIdx     = schema.getFieldIndex( DataCacheConstants.isDelField );
			delta_priorityIdx  = schema.getFieldIndex( DataCacheConstants.priorityField );
			delta_dataIdx      = schema.getFieldIndex( DataCacheConstants.dataField );
			dataSchema         = schema.getField( delta_dataIdx ).getSchema();
			
			// Update schema for consumer
			consumer.receiveDataCacheSchema( this.dataSchema );
		}
	}

	// This may need some re-start capabilities around it -- start request again and again!
	private void startDataCache() {

		// Start consuming the deltas
		DataCacheStreamBaseClient.getClient().subscribeToStream( path + "." + DataCacheConstants.DeltaStream, this );
		
		// Start consuming snapshots
		DataCacheStreamBaseClient.getClient().subscribeToStream( path + "." + DataCacheConstants.StateTableStream, this );
		
		// Set state to INITIALIZING
		this.cState = State.STATE_INITIALIZING;
		
		// Initalize request tuple
		try {
			this.requestTuple.setInt( req_requestIdIdx, Math.abs( rand.nextInt() ) );
		} catch ( TupleException e ) {
			System.out.println(" Fatal tuple exception, should never happen: " + e.toString() );
		}

		// Send off our request
		DataCacheStreamBaseClient.getClient().enqueueTuple( path + "." + DataCacheConstants.RequestAllStateStream, requestTuple );
	}
	
	@Override
	public void consumeTuple( Tuple tuple, String streamName ) {

		try {
			// This is where the state management needs to go!
			if ( streamName.equals( DataCacheConstants.StateTableStream ) ) {
	
				// Get requestIds!
				int requestId = tuple.getInt( state_requestIdIdx );
				int sentreqId = this.requestTuple.getInt( req_requestIdIdx );
				
				// Switch to loading if we were initializing and received token
				if ( cState == State.STATE_INITIALIZING && requestId == sentreqId ) {
					cState = State.STATE_LOADING;
					return;
				}
				
				// If loading and received a state update, then apply to model
				if ( cState == State.STATE_LOADING && requestId == sentreqId ) {
					applyTupleToModel( tuple.getString( state_keyFieldIdx ),
							tuple.getLong( state_priorityIdx ), false,
							tuple.getTuple( state_dataIdx ) );
					return;
				}
				
				// If loading and received end token, then switch to active and purge
				if ( cState == State.STATE_LOADING && requestId == ( -1 * sentreqId  ) ) {
					cState = State.STATE_ACTIVE;
					
					/* Before purging, let's delete elements that are marked for deletion */
					Iterator< Map.Entry< String, IDataCacheConsumer.DataObj > > iter = data.entrySet().iterator();
					while ( iter.hasNext() ) {
						Map.Entry< String, IDataCacheConsumer.DataObj > d = iter.next();
						if ( d.getValue().isDel )
							iter.remove();
					}
					
					/* Purge model to the downstream system */
					purgeModel();
					return;
				}
			}
	
			else if ( streamName.equals( DataCacheConstants.DeltaStream ) ) {
				
				// Simply update the model!
				applyTupleToModel( tuple.getString( delta_keyFieldIdx ),
						tuple.getLong( delta_priorityIdx ), tuple.getBoolean( delta_isDelIdx ),
						tuple.getTuple( delta_dataIdx ) );
			}
		}
		catch (TupleException e) {
			System.out.println("Fatal error parsing tuples, should never happen: " + e.toString() );
			e.printStackTrace();
		}
	}

	// Apply a delta (whether original state or subsequent delta) to our internal data cache
	private void applyTupleToModel( String key, long priority, boolean isDel, Tuple tuple) {

		// If the data is already in our cached data model and more up to date, then skip it
		if ( data.containsKey( key ) && data.get( key ).priority > priority )
			return;
		
		// As this is not in our model or more up to do, then just set it
		IDataCacheConsumer.DataObj dd  = new IDataCacheConsumer.DataObj();
		dd.key      = key;
		dd.priority = priority;
		dd.data     = tuple;
		dd.isDel    = isDel;
		data.put( key, dd );
		return;
	}

	// Purge our data model if it isn't empty, i.e., there are things to push downstream
	// AND we are active, then purge the model
	@Override
	public void notifyComplete() {
		if ( !data.isEmpty() && cState == State.STATE_ACTIVE )
			purgeModel();
	}

	private void purgeModel() {
		
		// Generate an arra of changes
		IDataCacheConsumer.DataObj out[] = data.values().toArray( new IDataCacheConsumer.DataObj[ data.size() ] );

		// Send off cache to downstream consumers
		consumer.receiveDataCacheUpdate( out );
		
		// Clear the cache for receiving new updates
		data.clear();
	}
}
