package com.streambase.contrib.datacache;

import java.util.Iterator;

import com.streambase.sb.EntityType;
import com.streambase.sb.StreamBaseException;
import com.streambase.sb.StreamProperties;
import com.streambase.sb.Tuple;
import com.streambase.sb.client.DequeueResult;
import com.streambase.sb.client.StreamBaseClient;
import com.streambase.sb.client.StreamBaseURI;
import com.streambase.sb.client.URIException;

/**
 * Wrapper over a {@link StreamBaseClient}, providing connectivity to the data cache framework.
 * Users of the data cache framework should connect to a streambase server via {@link #initClient(String[])}
 * prior to using any data cache service.
 * @author  mark.scannell@streambase.com
 */
public class DataCacheStreamBaseClient extends Thread {
	
	// before 7.0 this constant was StreamBaseClient.FULLY_QUALIFIED_NAMES
	// after 7.0 the constant is StreamBaseClient.ListEntityFlags.FULLY_QUALIFIED_NAMES
	private static final int FULLY_QUALIFIED_NAMES_CONSTANT = 1;
	
	/**
	 * Internal framework interface used by {@link DataCacheService}.
	 * Not intended to be implemented by framework users.
	 */
	public static interface IStreamConsumer {
		
		// Consume the schema for the subscription request
		public void consumeSchema( com.streambase.sb.Schema schema, String stream );
		
		// Consume a tuple for the subscription request
		public void consumeTuple( com.streambase.sb.Tuple tuple, String stream );
		
		// Be notified that a read loop has just been finished
		public void notifyComplete();
	}
	
	private static DataCacheStreamBaseClient singleton = null;
	
	/*package*/ static DataCacheStreamBaseClient getClient() {
		return singleton;
	}

	/**
	 * Initialize connectivity to a StreamBase server. Connects to the default
	 * streambase server (see {@link StreamBaseURI#fromEnvironment()}) if called with null,
	 * otherwise expects a StreamBase URI string as the first argument
	 * @param args an argument list such as that provided by Java via the <code>main</code> method, or null to connect to the default server
	 */
	public static void initClient( String args[] ) {
		if ( singleton == null )
			singleton = new DataCacheStreamBaseClient( args );
	}

	private StreamBaseClient sbc;
	
	// Initialization
	private DataCacheStreamBaseClient( String args[] ) {
		consumers = new java.util.HashMap<String, java.util.HashSet < IStreamConsumer > >();
		
		StreamBaseURI URI = null;
		try{
			if ( args.length >= 1 ) {
				URI = new StreamBaseURI(args[0]);
			} else {
				URI = StreamBaseURI.fromEnvironment();
			}
			System.out.println( "Connecting to StreamBase at " + URI.toString() );
			sbc = new StreamBaseClient( URI );
			sbc.enableHeartbeating();
			
			// Start background thread!
			startThread();
		}
		catch( URIException e ) {
			System.err.println( "Failed parsing URI: " + e.toString() );
		}
		catch ( StreamBaseException e ) {
			System.err.println( "Failed connecting to " + URI + ": " + e.toString() );
		}
	}
	
	// Start background thread that periodically calls tryDequeue on the client
	// and dispatches everything.
	private void startThread() {
		Runnable t = new java.lang.Runnable() {
			public void run() {
				while ( true ) {
					tryDequeue();
					try {
						Thread.sleep(50);
					} catch ( Exception e) {
					 /* */
					}
				}
			}
		};
		new Thread(t).start();
	}

	/*
	 * The following methods are all public accessing to the StreamBaseClient and require synchronization
	 */
	
	private java.util.HashMap < String, java.util.HashSet < IStreamConsumer > > consumers;
	synchronized /*package*/ void subscribeToStream( String streamName, IStreamConsumer cons ) {

		// Send them the schema that they've just subscribed to
		try {
			cons.consumeSchema( sbc.getSchemaForStream( streamName ), streamName );
		}
		catch ( StreamBaseException e ) {
			System.err.println( "Failed getting the schema for " + streamName );
		}
	
		// If not subscribed to this stream yet, then do so!
		if ( !consumers.containsKey( streamName ) ) {
			try {
				sbc.subscribe( streamName );
			} catch (StreamBaseException e) {
				System.err.println("StreamBaseException on stream subscribe: " + e.getMessage());
				e.printStackTrace();
			}
			consumers.put( streamName, new java.util.HashSet<IStreamConsumer>() );
		}

		// Add in this particular consumer
		consumers.get( streamName ).add( cons );	
	}
	
	synchronized /*package*/ java.util.Set< StreamProperties > getInputStreams() {
		try {
			java.util.Set< StreamProperties > i = new java.util.HashSet< StreamProperties >();
			for ( String c : sbc.listEntities( EntityType.CONTAINER ) ) {
				for ( String s : sbc.listEntities( c + ".stream", FULLY_QUALIFIED_NAMES_CONSTANT ) ) {
					StreamProperties sp = sbc.getStreamProperties( s );
					if ( sp.isInputStream() )
						i.add( sbc.getStreamProperties( s ) );
				}
			}
			return i;
		} catch ( StreamBaseException e ) {
			return null;
		}
	}

	synchronized /*package*/ java.util.Set< StreamProperties > getOutputStreams() {
		try {
			java.util.Set< StreamProperties > o = new java.util.HashSet< StreamProperties >();
			for ( String c : sbc.listEntities( EntityType.CONTAINER ) ) {
				for ( String s : sbc.listEntities( c + ".stream", FULLY_QUALIFIED_NAMES_CONSTANT ) ) {
					StreamProperties sp = sbc.getStreamProperties( s );
					if ( sp.isOutputStream() )
						o.add( sbc.getStreamProperties( s ) );
				}
			}
			return o;
		} catch ( StreamBaseException e ) {
			return null;
		}
	}

	synchronized /*package*/ void enqueueTuple( String stream, Tuple data ) {
		try {
			sbc.enqueue( stream, data);
		} catch ( StreamBaseException e ) {
			System.err.println("Error enqueueing tuple: " + e.toString() );
			e.printStackTrace();
		}
	}

	/*
	 * End of public functions
	 */
	
	/* Private internal thread that does the dequeuing from a separate thread - thus requiring synchronization */
	synchronized private void tryDequeue() {
		try {
			DequeueResult dr = null;
			
			// If we can dequeue, then do it!
			while ( sbc.canDequeue() && (dr = sbc.dequeue()) != null )
				dispatchResult( dr );
		} catch (StreamBaseException e) {
			System.err.println("StreamBaseException during dequeue: "
					+ e.getMessage());
			e.printStackTrace();
		}
	}
	
	private void dispatchResult( DequeueResult dr ) {
		
		java.util.HashSet< IStreamConsumer > toNotify = new java.util.HashSet< IStreamConsumer >();
		Iterator<Tuple> data = dr.iterator();
		while (data.hasNext()) {
			Tuple t = data.next();
			
			java.util.Set< IStreamConsumer > s = consumers.get( dr.getStreamProperties().getQualifiedName() );
			if ( s != null ) {
				for ( IStreamConsumer isc : s ) {
					isc.consumeTuple( t, dr.getStreamName() );
					toNotify.add( isc );
				}
			}

			s = consumers.get( dr.getStreamName() );
			if ( s != null ) {
				for ( IStreamConsumer isc : s ) {
					isc.consumeTuple( t, dr.getStreamName() );
					toNotify.add( isc );
				}
			}
		}

		// this needs to be a bit more precise! only for those who wanted this stream..
		for ( IStreamConsumer isc : toNotify )
			isc.notifyComplete();
	}
}
