package com.streambase.contrib.datacache;

import javax.swing.SwingUtilities;
import javax.swing.table.TableModel;

import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.swt.widgets.Display;

import com.streambase.contrib.datacache.gui.swing.CacheTableModel;
import com.streambase.contrib.datacache.gui.swing.DataCacheSwingMain;
import com.streambase.contrib.datacache.gui.swt.DataCacheDataBindingMain;
import com.streambase.contrib.datacache.gui.swt.DataCacheJFaceMain;
import com.streambase.contrib.datacache.gui.swt.DataCacheJFaceMain.DataCacheTableContentProvider;
import com.streambase.sb.Schema;
import com.streambase.sb.Tuple;

/**
 * Users of the Data Cache framework should provide an implementation of this to {@link DataCacheService#startDataCache(String, IDataCacheConsumer)}
 * in order to receive the query table data being maintained by the framework.
 * <p/>
 * The framework will call {@link #receiveDataCacheSchema(Schema)} at least once with information
 * about the structure of the data. Then, {@link #receiveDataCacheUpdate(DataObj[])} will be called
 * periodically to update this cache consumer with new data events (in the form of an array of {@link DataObj}).
 * <p/>
 * The framework provides a couple of examples of GUI data cache consumers:
 * <ul>
 * <li> {@link CacheTableModel} is a consumer that maps updates to a Swing {@link TableModel}</li>
 * <li> {@link DataCacheTableContentProvider} is a consumer that maps updates to an Eclipse JFace {@link IStructuredContentProvider} </li>
 * </ul>
 * @see DataCacheSwingMain
 * @see DataCacheJFaceMain
 * @see DataCacheDataBindingMain
 * 
 * @author mark.scannell@streambase.com
 */
public interface IDataCacheConsumer {
	
	/**
	 * Cache updates are communicated using this class. Instances may hold on to
	 * instances provided by the framework freely.
	 */
	public class DataObj {
		/**
		 * The StreamBase {@link Tuple} representing this data row
		 */
		public Tuple   data;
		/**
		 * The "key" value for this data cache object, as set by the StreamBase application author
		 */
		public String  key;
		/**
		 * When this boolean is set to <code>true</code>, the Tuple with key {@link #key} has been deleted
		 */
		public boolean isDel;
		/**
		 * A priority value that can be used to determine stale events (can be used as a tie-breaker when two
		 * or more {@link DataObj} instances refer to the same key in the same update event)
		 */
		public long    priority;
	};

	/**
	 * Receive a set of updates to the data cache. The first time this is called, implementors should expect
	 * a full table snapshot. Subsequent calls receive updates. Note that the framework <em>compresses</em> updates,
	 * that is, implementors will receive net-effect updates if the data changed more than once between refresh intervals.
	 * The argument object may be used freely by the receiver, as the framework will not re-use it.
	 * <p/>
	 * Typical usage by a GUI framework consuming Data Cache updates, is to dispatch widget updates in their appropriate
	 * event thread (using {@link SwingUtilities} for Swing, or {@link Display} for SWT) in response to this call.
	 * @param upd update events
	 * @see DataObj
	 */
	public void receiveDataCacheUpdate( DataObj upd[] );
	
	/**
	 * Receive the data schema for this data cache. This schema may be used to setup a user interface
	 * in preparation for receiving data updates.
	 * <br/>
	 * The framework will always calls this prior to the first call of {@link #receiveDataCacheUpdate(DataObj[])}.
	 * @param schema {@link Schema} associated with the DataCache this provides updates for  
	 */
	public void receiveDataCacheSchema( Schema schema );
}
