package com.streambase.contrib.datacache.gui.swing;

import java.util.Collections;
import java.util.Comparator;
import java.util.Vector;

import javax.swing.SwingUtilities;
import javax.swing.event.TableModelListener;
import javax.swing.table.AbstractTableModel;

import com.streambase.contrib.datacache.DataCacheService;
import com.streambase.contrib.datacache.IDataCacheConsumer;
import com.streambase.contrib.datacache.IDataCacheConsumer.DataObj;
import com.streambase.sb.DataType;
import com.streambase.sb.Schema;

@SuppressWarnings("serial")
public class CacheTableModel extends AbstractTableModel implements IDataCacheConsumer, java.io.Serializable {

	// Sorting in the vector for the keys
	private class dataComparator implements Comparator<IDataCacheConsumer.DataObj> {
		@Override
		public int compare(IDataCacheConsumer.DataObj arg0, IDataCacheConsumer.DataObj arg1) {
			try {
				return (arg0).key.compareTo( (arg1).key );
			} catch ( Exception e ) {
				return 0;
			}
		}
	}
	
	// Local data cache
	private Vector< IDataCacheConsumer.DataObj > data;
	private dataComparator cmp;
	private com.streambase.sb.DataType[] colTypes;
	private String[] colNames;
	
	// Initialize  (won't actually start until someone starts listening to data!)
	public CacheTableModel( String cacheName ) {
		cmp = new dataComparator();
		this.data = new Vector< IDataCacheConsumer.DataObj >();
		DataCacheService.startDataCache( cacheName, this );
	}

	@Override
	public int getColumnCount() {
		if ( colTypes != null ) 
			return colTypes.length;
		else
			return 0;
	}

	@Override
	public Class<?> getColumnClass(int arg0) {
		if ( colTypes != null ) {
			if ( colTypes[arg0] == DataType.LIST || colTypes[arg0] == DataType.TUPLE ) {
				return String.class;
			} else {
				return colTypes[arg0].getClassForType();
			}
		}
		else
			return null;
	}

	@Override
	public boolean isCellEditable(int arg0, int arg1) {
		return false;
	}

	@Override
	public String getColumnName(int arg0) {
		if ( colNames != null )
			return colNames[arg0];
		else
			return null;
	}

	@Override
	public int getRowCount() {
		try {
			return data.size();
		} catch ( Exception e ) {
			System.out.println("Exception getting row count: " + e.toString() );
		}
		return 0;
	}

	@Override
	public Object getValueAt(int arg0, int arg1) {
		try {
			if ( colTypes[ arg1 ] == DataType.LIST || colTypes[ arg1 ] == DataType.TUPLE ) {
				return data.get( arg0 ).data.getField( arg1 ).toString();
			} else {
				return data.get( arg0 ).data.getField( arg1 );
			}
		} catch ( com.streambase.sb.TupleException e ) {
			System.out.println("GetValueAt Exception: " + e.toString() );
			return null;
		}
	}

	@Override
	public void addTableModelListener( TableModelListener l ) {
		super.addTableModelListener( l );
	}

	@Override
	public void removeTableModelListener( TableModelListener l ) {
		super.removeTableModelListener( l );
	}

	// Dispatch until later with the data -- this is critical!
	@Override
	public void receiveDataCacheUpdate(DataObj[] upd) {
		final DataObj[] myUpd = upd;
		SwingUtilities.invokeLater( new Runnable() {
			@Override
			public void run() {
				// Update model with the data -- when Swing gets a chance to catch up!
				try {
					updateModel( myUpd );
				} catch ( Exception e ) {
					System.out.println("Exception updating model: " + e.toString() );
				}
			}
		});
	}
	
	// Update (in UI thread) with the incoming data
	private void updateModel( DataObj[] upd ) {
		
		// If empty, then load and then insert, sort, and indicate entire table changed!
		if ( data.size() == 0 ) {
			data.setSize( upd.length );
			for ( int i = 0; i < upd.length; i ++ ) {
				data.set( i, upd[ i ] );
			}
			Collections.sort( data, cmp );
			this.fireTableDataChanged();
			return;
		}

		// Not empty -- then process each individually and merge in!
		for ( int i = 0; i < upd.length; i ++ ) {
			
			int idx = Collections.binarySearch( data, upd[i], cmp );
			
			// If found -- then update or delete
			if ( idx >= 0 ) {

				// If existing has higher priority (this shouldn't happen!), then skip
				if ( data.get( idx ).priority > upd[i].priority ) {
					System.out.println("Higher priority detected.. shouldn't happen");
				}
				
				// Removed row - delete row and indicate fire deletion!
				else if ( upd[i].isDel ) {
					data.remove( idx );
					this.fireTableRowsDeleted( idx, idx );
				}
			
				// Otherwise this is an update
				else {
					data.set( idx, upd[i] );
					this.fireTableRowsUpdated( idx, idx );
				}
			}
			
			// Otherwise this is an insert
			else {
				int insertRow = -1 * ( idx + 1 );
				data.insertElementAt( upd[i], insertRow );
				this.fireTableRowsInserted( insertRow, insertRow );
			}
		}
	}

	// Capture schema (for headers and column types)
	@Override
	public void receiveDataCacheSchema(Schema schema) {
		colTypes = schema.getFieldTypes();
		colNames = schema.getFieldNames();
	}
}
