package com.streambase.contrib.datacache.gui.swt;

import org.eclipse.jface.viewers.ColumnViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.viewers.ViewerComparator;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;

/**
 * Usage:
 * <ul>
 * <li>Construct and maintain a new ColumnViewerSorter for the column you want to start sorting on</li>
 * <li>For each other column, create a new ColumnViewerSorter, with no need for a local reference</li>
 * <li>call {@link #setSorter(ColumnViewerSorter, int)} on the sole instance you held on to, in order to set the starting sort direction  
 * </ul>
 * Note that this sorter always responds <code>true</code> to {@link #isSorterProperty(Object, String)}
 * @author eddie
 * @see http://dev.eclipse.org/viewcvs/index.cgi/org.eclipse.jface.snippets/Eclipse%20JFace%20Snippets/org/eclipse/jface/snippets/viewers/Snippet040TableViewerSorting.java?view=markup
 */
public abstract class ColumnViewerSorter extends ViewerComparator {
	public static final int ASC = 1;
	public static final int NONE = 0;
	public static final int DESC = -1;
	private boolean showSortIndicator = true;

	private TableViewerColumn myColumn;
	private ColumnViewer myViewer;
	private int myDirection = NONE;

	@Override
	public boolean isSorterProperty(Object element, String property) {
		return true; // for any change, re-sort
	}

	public ColumnViewerSorter(ColumnViewer viewer, TableViewerColumn column) {
		this(viewer, column, true);
	}

	public ColumnViewerSorter(ColumnViewer viewer, TableViewerColumn column, boolean showSortIndicator) {
	    this.showSortIndicator = showSortIndicator;
	    this.myColumn = column;
        this.myViewer = viewer;
        this.myColumn.getColumn().addSelectionListener(new SelectionAdapter() {
            @Override
            public void widgetSelected(SelectionEvent e) {
                if (myViewer.getComparator() != null ) {
                    if (myViewer.getComparator() == ColumnViewerSorter.this ) {
                        int tdirection = ColumnViewerSorter.this.myDirection;

                        if( tdirection == ASC ) {
                            setSorter(ColumnViewerSorter.this, DESC);
                        } else if( tdirection == DESC ) {
                            setSorter(ColumnViewerSorter.this, NONE);
                        }
                    } else {
                        setSorter(ColumnViewerSorter.this, ASC);
                    }
                } else {
                    setSorter(ColumnViewerSorter.this, ASC);
                }

            }
        });
	}
	/**
	 * Must be called in the UI thread, as a refresh of the table may take place
	 * @param sorter the sorter
	 * @param direction the direction to sort in
	 */
	public void setSorter(ColumnViewerSorter sorter, int direction) {
		if( direction == NONE ) {
			myColumn.getColumn().getParent().setSortColumn(null);
			myColumn.getColumn().getParent().setSortDirection(SWT.NONE);
			myViewer.setComparator(null);
		} else {
			myColumn.getColumn().getParent().setSortColumn(myColumn.getColumn());
			sorter.myDirection = direction;
            
			if (showSortIndicator) {
                if (direction == ASC) {
                    myColumn.getColumn().getParent().setSortDirection(SWT.UP);
                } else {
                    myColumn.getColumn().getParent().setSortDirection(SWT.DOWN);
                }
            }

			if( myViewer.getComparator() == sorter ) {
				myViewer.refresh();
			} else {
				myViewer.setComparator(sorter);
			}
		}
	}

	@Override
	public int compare(Viewer viewer, Object e1, Object e2) {
		return myDirection * doCompare(viewer, e1, e2);
	}

	protected abstract int doCompare(Viewer viewer, Object e1, Object e2);
}