package com.streambase.contrib.datacache.gui.swt;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.eclipse.core.databinding.observable.Realm;
import org.eclipse.core.databinding.observable.list.WritableList;
import org.eclipse.core.databinding.property.INativePropertyListener;
import org.eclipse.core.databinding.property.ISimplePropertyListener;
import org.eclipse.core.databinding.property.value.SimpleValueProperty;
import org.eclipse.jface.databinding.swt.SWTObservables;
import org.eclipse.jface.databinding.viewers.ViewerSupport;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.TabItem;
import org.eclipse.swt.widgets.TableColumn;

import com.streambase.contrib.datacache.DataCacheService;
import com.streambase.contrib.datacache.IDataCacheConsumer;
import com.streambase.contrib.datacache.gui.SharedUtilities;
import com.streambase.sb.Schema;



/**
 * Sample SWT/JFace UI using Data Binding, capable of displaying a table for any DataCache currently running
 * in a StreamBase application.
 * 
 * @author eddie@streambase.com
 */
public class DataCacheDataBindingMain extends DataCacheJFaceMain {

	public DataCacheDataBindingMain(Shell shell) {
		super(shell);
	}

	@Override
	protected void createDataCacheViewer(TabItem tab, String modulePath) {
		Composite c = new Composite(tab.getParent(), SWT.NONE);
		// note, unlike the base JFace example, we set a fill layout here and not a TableColumnLayout
		//  since we are not in charge of creating the columns (data binding is)
		c.setLayout(new FillLayout());
		final TableViewer viewer = new TableViewer(c, SWT.SINGLE | SWT.H_SCROLL | SWT.V_SCROLL | SWT.FULL_SELECTION);
		tab.setControl(c);
		
		viewer.getTable().setHeaderVisible(true);
		
		// begin the data cache -- note that starting it here in the UI thread means
		//  the schema will be provided back in the UI thread as well 
		DataCacheService.startDataCache(modulePath, new IDataCacheConsumer() {
			private WritableList model = new WritableList(SWTObservables.getRealm(viewer.getControl().getDisplay()), (Collection<DataObj>) new ArrayList<DataObj>(), DataObj.class);
			
			@Override
			public void receiveDataCacheUpdate(final DataObj[] updatedElements) {
				if (viewer.getControl().isDisposed()) return;
				SWTObservables.getRealm(viewer.getControl().getDisplay()).asyncExec(new Runnable() {
					
					@Override
					public void run() {
						updateModel(updatedElements);
					}
				});
			}

			private void updateModel(DataObj[] updatedElements) {
				// TODO XXX Revisit this binary search/model-is-sorted-or-not thing going on
				//           model shouldn't be sorted imho, but key should be used for "Set" behaviour...
				
				// Not empty -- then process each individually and merge in!
				for ( int i = 0; i < updatedElements.length; i ++ ) {
					int idx = -1;
					
					for (int j = 0; j < model.size(); ++j) {
						DataObj obj = (DataObj) model.get(j);
						if (obj.key.equals(updatedElements[i].key)) {
							idx = j;
							break;
						}
					}
					
					// If found -- then update or delete
					if ( idx >= 0 ) {

						// If existing has higher priority (this shouldn't happen!), then skip
						final DataObj existingElement = (DataObj) model.get(idx);
						if ( existingElement.priority > updatedElements[i].priority ) {
							System.err.println("Higher priority detected.. shouldn't happen");
						}
						
						// Removed row - delete row
						else if ( updatedElements[i].isDel ) {
							model.remove( idx );							
						}
					
						// Otherwise this is an update
						else {
							// so data binding sees the change, we use set to the new data object, rather
							// than updating the object itself
							model.set(idx, updatedElements[i]);
						}
					}
					
					// Otherwise this is an insert
					else {
						model.add(updatedElements[i]);
					}
				}
			}
			
			@Override
			public void receiveDataCacheSchema(final Schema schema) {
				List<SimpleValueProperty> valueProperties = new ArrayList<SimpleValueProperty>();
				for (final Schema.Field field : schema.fields()) {
					TableColumn col = new TableColumn(viewer.getTable(), SWT.LEFT);
					col.setText(field.getName());
					col.pack();
					
					valueProperties.add(new SimpleValueProperty() {
						
						@Override
						public Object getValueType() {
							// XXX investigate need for this
							return field.getCompleteDataType();
						}
						
						@Override
						protected void doSetValue(Object source, Object value) {
							// no-op: not supported
						}
						
						@Override
						protected Object doGetValue(Object source) {
							return ((DataObj) source).data.getField(field);
						}
						
						@Override
						public INativePropertyListener adaptListener(
								ISimplePropertyListener listener) {
							// not supported
							return null;
						}
					});
				}
				
				//   wait until "DataObj" is either beany, or we go ahead and write this...
				ViewerSupport.bind(viewer, model, valueProperties.toArray(new SimpleValueProperty[0]));
			}
		});
	}
	
	public static void main(String[] args) {
		// Initialize
		SharedUtilities.initializeServerConnectivity(args);
		
		// Kick off UI application window under a realm mapped to the SWT UI Thread
		final Display display = new Display();
		Realm.runWithDefault(SWTObservables.getRealm(display), new Runnable() {
			public void run() {
				Shell shell = new Shell(display);
				new DataCacheDataBindingMain(shell).run();
			}
		});

		
	}
}
