package com.streambase.contrib.datacache.gui.swt;

import java.text.Collator;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;

import org.eclipse.jface.layout.GridDataFactory;
import org.eclipse.jface.layout.GridLayoutFactory;
import org.eclipse.jface.layout.TableColumnLayout;
import org.eclipse.jface.viewers.ColumnLabelProvider;
import org.eclipse.jface.viewers.ColumnWeightData;
import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.ITableLabelProvider;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.window.ApplicationWindow;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.TabFolder;
import org.eclipse.swt.widgets.TabItem;

import com.streambase.contrib.datacache.DataCacheService;
import com.streambase.contrib.datacache.IDataCacheConsumer;
import com.streambase.contrib.datacache.IDataCacheConsumer.DataObj;
import com.streambase.contrib.datacache.gui.SharedUtilities;
import com.streambase.sb.Schema;
import com.streambase.sb.Tuple;
import com.streambase.sb.TupleException;
import com.streambase.sb.Tuple.TupleFormatter;

/**
 * Sample SWT/JFace UI capable of displaying a table for any DataCache currently running
 * in a StreamBase application.
 * 
 * @author eddie@streambase.com
 */
public class DataCacheJFaceMain extends ApplicationWindow {
	private static class DataObjComparator implements Comparator<IDataCacheConsumer.DataObj> {
		@Override
		public int compare(IDataCacheConsumer.DataObj arg0, IDataCacheConsumer.DataObj arg1) {
			try {
				return (arg0).key.compareTo( (arg1).key );
			} catch ( Exception e ) {
				return 0;
			}
		}
	}

	protected static DataObjComparator dataObjComparator = new DataObjComparator();


	/**
	 * our label providers, and sorter providers, will respond to all properties
	 */
	private static final String[] VIEWER_ALL_PROPERTIES = new String[] { "_ALL_" };

	/**
	 * Label provider for {@link DataObj}
	 */
	public static class DataCacheTableLabelProvider extends LabelProvider implements ITableLabelProvider {

		@Override
		public Image getColumnImage(Object element, int columnIndex) {
			return null;
		}

		@Override
		public String getColumnText(Object element, int columnIndex) {
			DataObj obj = (DataObj) element;
			try {
				return obj.data.getField(columnIndex).toString();
			} catch (TupleException e) {
				System.err.println(e);
				return "(error: " + e.getMessage() + ")";
			}
		}

	}

	/**
	 * A content provider for a data cache table. The input should be a String, corresponding to
	 * the data cache name to connect to. The content provider will register against the {@link DataCacheService}
	 * and begin pulling content from it.  
	 */
	public static class DataCacheTableContentProvider implements IStructuredContentProvider, IDataCacheConsumer {
		TupleFormatter DEFAULT_FORMATTER = new Tuple.DefaultTupleFormatter();
	    
		/**
		 * The model. Note, do *not* learn from this that content providers
		 * should actually have the model, they shouldn't! This is just for simplicity.
		 */
		private Map<String, DataObj> data;


		private TableViewer viewer;

		@Override
		public Object[] getElements(Object inputElement) {
			return data.values().toArray();
		}

		@Override
		public void dispose() {
		}

		@Override
		public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {
			if (newInput != null) { // when null, we are disposing it
				this.viewer = (TableViewer) viewer;
				// XXX this is a gross hack way of using a content provider, but gets the job done in this case
				//      namely, you should handle old/new inputs, and this content provider shouldn't actually
				//      OWN the model internally
				this.data = new HashMap<String, IDataCacheConsumer.DataObj>();		
				// begin the data cache -- note that starting it here in the UI thread means
				//  the schema will be provided back in the UI thread as well 
				DataCacheService.startDataCache((String) newInput, this);
			}
		}

		@Override
		public void receiveDataCacheSchema(final Schema schema) {
			// setup the table immediately (this is in the UI thread due to this being
			// called from the inputChanged method starting the data cache immediately 
			if (viewer.getControl().isDisposed()) return;
			if (viewer.getTable().getColumnCount() != 0) {
				System.err.println("Ignoring schema change (reconnect?)");
				return;
			}

			TableColumnLayout tcl = new TableColumnLayout();
			viewer.getTable().getParent().setLayout(tcl);
			ColumnViewerSorter startColSorter = null;

			for (final Schema.Field field : schema.fields()) {
				TableViewerColumn col = new TableViewerColumn(viewer, SWT.LEFT);
				col.getColumn().setText(field.getName());
				col.getColumn().setResizable(true);

				col.setLabelProvider(new ColumnLabelProvider() {
					@Override
					public String getText(Object element) {
						DataObj obj = (DataObj) element;
						return obj.data.getFieldAsString(field, DEFAULT_FORMATTER);
					}
				});

				tcl.setColumnData(col.getColumn(), new ColumnWeightData(100 / schema.getFieldCount()));
				// sorting support
				ColumnViewerSorter sorter = new ColumnViewerSorter(viewer, col) {

					@SuppressWarnings("unchecked")
					@Override
					protected int doCompare(Viewer viewer, Object e1, Object e2) {
						DataObj row1 = (DataObj) e1;
						DataObj row2 = (DataObj) e2;

						Object v1 = row1.data.getField(field);
						Object v2 = row2.data.getField(field);

						if (v1 instanceof Comparable<?> && v2 instanceof Comparable<?>) {
							return ((Comparable) v1).compareTo((Comparable)v2);
						}
						return Collator.getInstance().compare(
								row1.data.getFieldAsString(field, DEFAULT_FORMATTER),
								row2.data.getFieldAsString(field, DEFAULT_FORMATTER));
					}
				};
				if (startColSorter == null) {
					startColSorter = sorter;
					startColSorter.setSorter(startColSorter, ColumnViewerSorter.ASC);
				}
			}
			viewer.getTable().getParent().layout();
		}

		@Override
		public void receiveDataCacheUpdate(final DataObj[] updatedElements) {
			if (viewer.getControl().isDisposed()) return;

			if (data.size() == 0) {
				// they are all newly added items
				for (DataObj dataObj : updatedElements) {
					data.put(dataObj.key, dataObj);
				}
				// update UI
				viewer.getControl().getDisplay().asyncExec(new Runnable() {
					
					@Override
					public void run() {
						if (viewer.getControl().isDisposed()) return;
						viewer.add(updatedElements);
					}
				});
				return;
			}
			
			// these allow us to batch-update the UI after updating our model
			// XXX could prevent re-allocation by constructing once, and clearing each time
			final List<DataObj> deletes = new ArrayList<DataObj>();
			final List<DataObj> inserts = new ArrayList<DataObj>();
			final List<DataObj> updates = new ArrayList<DataObj>();
			
			// Not empty -- then process each individually and merge in!
			for ( int i = 0; i < updatedElements.length; ++i ) {
				
				boolean existing = data.containsKey(updatedElements[i].key);

				// If found -- then update or delete
				if ( existing ) {

					// If existing has higher priority (this shouldn't happen!), then skip
					final DataObj existingElement = data.get(updatedElements[i].key);
					
					if ( existingElement.priority > updatedElements[i].priority ) {
						System.err.println("Higher priority data detected.. shouldn't happen");
					}

					// Removed row - delete row
					else if ( updatedElements[i].isDel ) {
						DataObj old = data.remove( existingElement.key );
						deletes.add(old);
					}

					// Otherwise this is an update
					else {
						// do a copy to ensure we maintain an update in the model, and not a remove/add
						DataObj newElem = updatedElements[i];

						existingElement.data = newElem.data;
						existingElement.isDel = newElem.isDel;
						existingElement.key = newElem.key;
						existingElement.priority = newElem.priority;
						updates.add(existingElement);
					}
				}

				// Otherwise this is an insert
				else {							
					data.put(updatedElements[i].key, updatedElements[i]);
					inserts.add(updatedElements[i]);
				}
			}
			// UI update time
			if (deletes.size() > 0 || inserts.size() > 0 || updates.size() > 0) {
				viewer.getControl().getDisplay().asyncExec(new Runnable() {

					@Override
					public void run() {
						if (viewer.getControl().isDisposed()) return;

						if (deletes.size() > 0) viewer.remove(deletes.toArray());
						if (inserts.size() > 0) viewer.add(inserts.toArray());
						if (updates.size() > 0) viewer.update(updates.toArray(), VIEWER_ALL_PROPERTIES);
					}

				});
			}
		}

	}


	public DataCacheJFaceMain() {
		this(null);
	}

	public DataCacheJFaceMain(Shell shell) {
		super(shell);
	}

	@Override
	public boolean close() {
		boolean ret = super.close();
		System.exit(0);
		return ret;
	}

	public void run() {
		setBlockOnOpen(true);

		open();

		Display.getCurrent().dispose();
	}

	@Override
	protected Point getInitialSize() {
		return new Point(600, 500);
	}

	@Override
	protected Control createContents(Composite parent) {
		Composite composite = new Composite(parent, SWT.NONE);		
		SortedSet<String> dataCaches = new TreeSet<String>(DataCacheService.getDataCaches());
		
		final Combo dataCacheChooser = new Combo(composite, SWT.DROP_DOWN | SWT.READ_ONLY);
		dataCacheChooser.setItems(dataCaches.toArray(new String[0]));

		final TabFolder dataCacheTableFolder = new TabFolder(composite, SWT.TOP);
		GridDataFactory.fillDefaults().grab(true, true).applyTo(dataCacheTableFolder);


		dataCacheChooser.addSelectionListener(new SelectionAdapter() {

			@Override
			public void widgetSelected(SelectionEvent e) {
				showDataCache(dataCacheTableFolder, dataCacheChooser.getText());
			}
		});

		GridLayoutFactory.swtDefaults().generateLayout(composite);
		return composite;
	}

	private void showDataCache(TabFolder folder, String modulePath) {
		for (TabItem tab : folder.getItems()) {
			if (tab.getText().equals(modulePath)) {
				folder.setSelection(tab);
				return;
			}
		}

		TabItem tab = new TabItem(folder, SWT.NONE);
		tab.setText(modulePath);
		folder.setSelection(tab);
		
		createDataCacheViewer(tab, modulePath);		

	}

	protected void createDataCacheViewer(TabItem tab, String modulePath) {
		// layout on the tab's composite will be set later: see the viewer's content provider
		Composite c = new Composite(tab.getParent(), SWT.NONE);
		TableViewer tv = new TableViewer(c, SWT.SINGLE | SWT.H_SCROLL | SWT.V_SCROLL | SWT.FULL_SELECTION);
		tab.setControl(c);

		tv.getTable().setHeaderVisible(true);
		tv.getTable().setLinesVisible(true);
		tv.setContentProvider(new DataCacheTableContentProvider());
		// label providers are per column, see the content provider 
		tv.setInput(modulePath);
	}

	public static void main(String[] args) {
		// Initialize
		SharedUtilities.initializeServerConnectivity(args);
		
		// Kick off UI application window
		new DataCacheJFaceMain().run();
	}
}
